//****************************************************************************
// File: runlib.c
//
// Purpose: source file for RunLib library
//
// Functions: LibMain, WEP, Hook, ToolHelpNotifyProc, NodeAdd, NodeDelete,
//            NodeFind, SpawnAndWait, SpawnAndWaitIndirect
//
//
// Development Team: richhe
//
//
// Written by Microsoft Product Support Services, Windows Developer Support
// Copyright (c) 1992 Microsoft Corporation. All rights reserved.
//****************************************************************************

#include <windows.h>
#include <shellapi.h>
#include <toolhelp.h>
#include <malloc.h>
#include <memory.h>
#include "runlib.h"

HTASK  ghTask        = NULL;
LPNODE glpHead       = NULL;

//***********************************************************************
// Function: LibMain
//
// Purpose: C function called from DLL entry point. Loads RUNTASK.MOD
//          which is RUNLIB's private task to own callback proc
//
// Parameters: HINSTANCE hInst;   DLL instance handle
//             WORD wSeg;         DLL data segment selector
//             WORD cbHeapSize;   DLL initial heap size in bytes
//             LPSTR lpszCmdLine; DLL command line
//
//
// Returns: 1 is success, 0 fail DLL load
//
// Comments:
//
// History:  Date       Author        Reason
//
//****************************************************************************

int CALLBACK LibMain(HINSTANCE hInst, WORD wDataSeg, WORD cbHeapSize, LPSTR lpszCmdLine)
{
  HINSTANCE      hRunTask;
  CMDSHOW        CmdShow;
  PARAMETERBLOCK ParameterBlock;

  // zero init stack vars

  _fmemset(&ParameterBlock, 0, sizeof(PARAMETERBLOCK));
  _fmemset(&CmdShow, 0, sizeof(CMDSHOW));

  // setup and LoadModule the private task

  ParameterBlock.lpCmdLine          = "";
  ParameterBlock.lpCmdShow          = &CmdShow;
  ParameterBlock.lpCmdShow->wFirst  = 2;
  hRunTask = LoadModule("RUNTASK.MOD", &ParameterBlock);

  // must release control to system to let LoadModule run

  Yield();

  // if the application didn't load, fail the library load

  if(hRunTask > HINSTANCE_ERROR)
     return(1);
  else
     return(0);

}

//***********************************************************************
// Function: WEP
//
// Purpose: DLL exit procedure 
//
//
// Parameters: int nExitType; DLL shutdown code
//
//
//
//
//
// Returns: 1 
//
// Comments:
//
// History:  Date       Author        Reason
//
//****************************************************************************

int CALLBACK __export WEP(int nExitType)
{
   // if owner task around, unhook and tell it to quit

   if(ghTask != NULL)
   {
      NotifyUnRegister(ghTask);
      PostAppMessage(ghTask, WM_QUIT, 0, 0L);
   }
   return(1);
}

//***********************************************************************
// Function: Hook
//
// Purpose: inform DLL that private task successfully registered callback
//
//
// Parameters: none
//
//
//
//
//
// Returns: nothing
//
// Comments:
//
// History:  Date       Author        Reason
//
//****************************************************************************

void CALLBACK __export Hook()
{
   // save away private task handle

   ghTask = GetCurrentTask();
}

//***********************************************************************
// Function: ToolHelpNotifyProc
//
// Purpose: TOOLHELP.DLL NotifyRegister callback proc used to get 
//          NFY_STARTTASK and NFY_EXITTASK notifications
//
// Parameters: WORD wID;     notification code 
//             DWORD dwData; depends on wID
//
//
//
//
// Returns: FALSE to always chain on to next in callback list
//
// Comments:
//
// History:  Date       Author        Reason
//
//****************************************************************************

BOOL CALLBACK __export ToolHelpNotifyProc(WORD wID, DWORD dwData)
{
   LPNODE lpNode;

   switch(wID)
   {
      // task starting, look for empty list element and save away
      // if SpawnAndWaitIndirect did the exec'ing

      case NFY_STARTTASK:
         lpNode = NodeFind(0);
         if(lpNode != NULL)
            lpNode->hTask = GetCurrentTask();
         break;

      // task ending, clear element if one of our own
      // and save away return code

      case NFY_EXITTASK:
         lpNode = NodeFind(GetCurrentTask());
         if(lpNode != NULL)
         {
            lpNode->wCode = LOWORD(dwData);
            lpNode->hTask = 0;
         }
         break;

      default:
         break;
   }
   return(FALSE);
}

//***********************************************************************
// Function: SpawnAndWait
//
// Purpose: wrapper function for SpawnAndWaitIndirect callable from VB
//
//
// Parameters: HWND hwnd;           hwnd for ShellExecute 
//             LPCSTR lpszOp;       lpszOp for ShellExecute
//             LPCSTR lpszFile;     lpszFile for ShellExecute
//             LPCSTR lpszParams;   lpszParams for ShellExecute 
//             LPCSTR lpszDir;      lpszDir for ShellExecute
//             int fsCmdShow;       fsCmdShow for ShellExecute
//             LPMSGPROC lpMsgProc; callback proc to process messages
//
// Returns: HIWORD == hInstance of spawned application
//          LOWORD == return code of spawned application
//
// Comments:
//
// History:  Date       Author        Reason
//
//****************************************************************************

DWORD CALLBACK __export SpawnAndWait(HWND hwnd, LPCSTR lpszOp, LPCSTR lpszFile, LPCSTR lpszParams, LPCSTR lpszDir, int fsShowCmd)
{
   SPAWNWAIT sw;

   // load the parameters into struct and call real function

   sw.hwnd      = hwnd      ;       
   sw.lpszOp    = lpszOp    ;     
   sw.lpszFile  = lpszFile  ;   
   sw.lpszParams= lpszParams; 
   sw.lpszDir   = lpszDir   ;    
   sw.fsShowCmd = fsShowCmd ;  
   sw.lpmsgproc = NULL ;       // VB can't do message pump

   return(SpawnAndWaitIndirect(&sw));
}

//***********************************************************************
// Function: SpawnAndWaitIndirect
//
// Purpose: returns only when spawned application terminates for real
//          blocking on a program for syncronous spanwns
//
// Parameters: LPSPAWNWAIT lpSpawnWait; far reference to SPAWNWAIT struct
//
//
//
//
//
// Returns: HIWORD == hInstance of spawned application
//          LOWORD == return code of spawned application
//
// Comments:
//
// History:  Date       Author        Reason
//
//****************************************************************************

DWORD CALLBACK __export SpawnAndWaitIndirect(LPSPAWNWAIT lpSpawnWait)
{
   MSG         msg;
   LPNODE      lpNode;
   HINSTANCE   hChildInst = ((HINSTANCE) (-1));
   WORD        wCode = ((WORD) (-1));

   // if the private task is loaded and hooked into chain

   if(ghTask != NULL)
   {

   // allocate node in linked list and if successfull

      lpNode = NodeAdd();
      if(lpNode != NULL)
      {

   // spawn the program

         hChildInst = ShellExecute(lpSpawnWait->hwnd, lpSpawnWait->lpszOp,
                           lpSpawnWait->lpszFile, lpSpawnWait->lpszParams, 
                           lpSpawnWait->lpszDir, lpSpawnWait->fsShowCmd);

   // if its successfull, drop into the PeekMessage loop

         lpNode->wCode = wCode;
         if(hChildInst > HINSTANCE_ERROR)
         {
         
   // gets cleared by callback proc on NFY_EXITASK or error

            while(lpNode->hTask != NULL)
            {

   // remove the messages and process

               while(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
               {

   // quit is bad, so get out cleanly to let app process it

                  if(msg.message == WM_QUIT)
                  {
                     PostQuitMessage(msg.wParam);
                     lpNode->hTask = 0;
                     break;
                  }

   // no custom message handler, do standard processing

                  else if(lpSpawnWait->lpmsgproc == NULL)
                  {
                     TranslateMessage(&msg);
                     DispatchMessage(&msg);
                  }

   // call application provided message handler

                  else 
                     lpSpawnWait->lpmsgproc(&msg);
               }
            }
         }

   // save away return code and free linked list node

         wCode = lpNode->wCode;
         NodeDelete(lpNode);
      }
   }

   // return the return code

   return(MAKELONG(wCode,hChildInst));
}

//***********************************************************************
// Function: NodeAdd
//
// Purpose: allocates and adds a node to linked list of NODE's
//
//
// Parameters: none
//
//
//
//
//
// Returns: LPNODE of new element on sucess or NULL on fail
//
// Comments:
//
// History:  Date       Author        Reason
//
//****************************************************************************

LPNODE PASCAL NodeAdd(void)
{
   LPNODE lpNew  = NULL;

   // simple linked list add the new node at front of list

   lpNew = (LPNODE) _fmalloc(sizeof(NODE));
   if(lpNew != NULL)
   {
      _fmemset(lpNew, 0, sizeof(NODE));
      lpNew->lpNext = glpHead;
      glpHead = lpNew;
   }
   return(lpNew);
}

//***********************************************************************
// Function: NodeDelete
//
// Purpose: walk list, remove, and delete NODE from list
//
//
// Parameters: LPNODE lpNode; pointer to NODE struct created via NodeAdd
//
//
//
//
//
// Returns: nothing
//
// Comments:
//
// History:  Date       Author        Reason
//
//****************************************************************************

void PASCAL NodeDelete(LPNODE lpNode)
{
   LPNODE lpCurrent;
   LPNODE lpPrev;

   // simple linked list delete procedure

   lpCurrent = glpHead;
   while(lpCurrent != NULL)
   {
      lpPrev = lpCurrent;
      if(lpCurrent == lpNode)
      {
         lpPrev = lpCurrent->lpNext;
         _ffree(lpCurrent);
         lpCurrent = NULL;
      }
      else
         lpCurrent = lpCurrent->lpNext;
   }
}  

//***********************************************************************
// Function: NodeFind
//
// Purpose: find node with matching hTask value
//
//
// Parameters: HTASK hTask; task handle to find match in list
//
//
//
//
//
// Returns: LPNODE of node with matching hTask on sucess or NULL on fail
//
// Comments:
//
// History:  Date       Author        Reason
//
//****************************************************************************

LPNODE PASCAL NodeFind(HTASK hTask)
{
   LPNODE lpNode;

   // simple linked list search and find routine

   lpNode = glpHead;
   while(lpNode != NULL)
   {
      if(lpNode->hTask == hTask)
         break;
      lpNode = lpNode->lpNext;
   }
   return(lpNode);
}

